@@ -1,9 +1,9 @@
 use crate::{
-    command::{Actions, ProtoOpts},
     connection::Connection,
     ffi_util::IntoCString,
-    macros::{MacroMap, Stage},
-    reply::Reply,
+    macros::{Macros, Stage},
+    message::replies::Reply,
+    proto_util::{Actions, ProtoOpts},
 };
 use async_trait::async_trait;
 use std::{
@@ -107,50 +107,48 @@ impl From<io::Error> for ContextError {
     }
 }
 
+/// Callback context for the `negotiate` stage.
 pub struct NegotiateContext<T: Send> {
     pub data: Option<T>,
-
-    // note: there are no macros at this stage
-    pub reply: ReplyHandle,
-
+    pub reply: SmtpReply,
     pub requested_actions: Actions,
-    pub requested_proto_opts: ProtoOpts,
+    pub requested_opts: ProtoOpts,
     pub requested_macros: HashMap<Stage, CString>,
 }
 
 impl<T: Send> NegotiateContext<T> {
     pub(crate) fn new(
-        reply: ReplyHandle,
+        data: Option<T>,
+        reply: SmtpReply,
         requested_actions: Actions,
-        requested_proto_opts: ProtoOpts,
+        requested_opts: ProtoOpts,
     ) -> Self {
         Self {
-            data: None,  // always `None` initially
+            data,
             reply,
             requested_actions,
-            requested_proto_opts,
+            requested_opts,
             requested_macros: HashMap::new(),
         }
     }
 }
 
+/// Callback context.
 pub struct Context<T: Send> {
     pub data: Option<T>,
 
     // TODO if we provide a &mut reference to the context then this is all open,
     // and can be replaced by users; must somehow be read-only/protected/uninstantiable
-    pub macros: MacroMap,
-    // TODO alternatively, we could also expose ErrorReply directly -- then test code
-    // can mock it as `&mut Option<ErrorReply>`?
-    pub reply: ReplyHandle,
+    pub macros: Macros,
+    pub reply: SmtpReply,
 }
 
 impl<T: Send> Context<T> {
     pub(crate) fn new() -> Self {
         Self {
             data: None,
-            macros: MacroMap::new(),
-            reply: ReplyHandle::new(),
+            macros: Macros::new(),
+            reply: SmtpReply::new(),
         }
     }
 
@@ -166,33 +164,42 @@ impl<T: Send> Context<T> {
         self.macros.insert(stage, entries);
     }
 
-    pub(crate) fn restore_from_eom(&mut self, cx: EomContext<T>) {
-        self.data = cx.data;
-        self.reply = cx.reply;
+    pub(crate) fn restore(&mut self, data: Option<T>, reply: SmtpReply) {
+        self.data = data;
+        self.reply = reply;
+    }
+
+    pub(crate) fn take_reply_if<F>(&mut self, f: F) -> Option<Reply>
+    where
+        F: FnMut(&ReplyCode) -> bool,
+    {
+        self.reply
+            .take_reply_if_match(f)
+            .map(|reply| Reply::ReplyCode { reply })
     }
 }
 
+/// Callback context for the `eom` stage.
 pub struct EomContext<T: Send> {
     pub data: Option<T>,
-
-    pub macros: MacroMap,
-    pub reply: ReplyHandle,
-    pub actions: EomContextActions,
+    pub macros: Macros,
+    pub reply: SmtpReply,
+    pub actions: EomActions,
 }
 
 impl<T: Send> EomContext<T> {
     pub fn new(
         conn: Connection,
         data: Option<T>,
-        macros: MacroMap,
-        reply: ReplyHandle,
+        macros: Macros,
+        reply: SmtpReply,
         available_actions: Actions,
     ) -> Self {
         Self {
             data,
             macros,
             reply,
-            actions: EomContextActions {
+            actions: EomActions {
                 conn,
                 available_actions,
             },
@@ -200,33 +207,39 @@ impl<T: Send> EomContext<T> {
     }
 }
 
+// No Default, Clone, must be uninstantiable by users.
+/// Current custom SMTP reply.
 #[derive(Debug)]
-pub struct ReplyHandle {
-    reply: Option<ErrorReply>,
+pub struct SmtpReply {
+    error_reply: Option<ErrorReply>,
 }
 
-impl ReplyHandle {
+impl SmtpReply {
     pub(crate) fn new() -> Self {
-        Self { reply: None }
+        Self { error_reply: None }
     }
 
     pub(crate) fn duplicate(&self) -> Self {
-        Self { reply: self.reply.clone() }
+        Self {
+            error_reply: self.error_reply.clone(),
+        }
     }
 
     // TODO revisit
-    pub(crate) fn take_reply_if_init_eq(&mut self, initial: u8) -> Option<CString> {
-        if let Some(reply) = self.reply.as_ref() {
-            if reply.rcode.as_ref().as_bytes()[0] == initial {
-                let r = self.reply.take().unwrap();
-                return Some(r.make_error_reply());
+    pub(crate) fn take_reply_if_match<F>(&mut self, mut f: F) -> Option<CString>
+    where
+        F: FnMut(&ReplyCode) -> bool,
+    {
+        if let Some(reply) = self.error_reply.as_ref() {
+            if f(&reply.rcode) {
+                return self.error_reply.take().map(|r| r.make_error_reply());
             }
         }
         None
     }
 }
 
-impl SetErrorReply for ReplyHandle {
+impl SetErrorReply for SmtpReply {
     fn set_error_reply<I, T>(
         &mut self,
         rcode: &str,
@@ -245,21 +258,25 @@ impl SetErrorReply for ReplyHandle {
             .map_err(|_| ContextError::InvalidEnhancedStatusCode)?;
 
         let mut message_lines = Vec::new();
+
         for (i, line) in message.into_iter().enumerate() {
             if i >= 32 {
                 return Err(ContextError::InvalidReplyText);
             }
+
             let line = line.into_c_string();
+
             if line.as_bytes().len() > 980 {
                 return Err(ContextError::InvalidReplyText);
             }
             if line.as_bytes().iter().any(|&b| matches!(b, b'\r' | b'\n')) {
                 return Err(ContextError::InvalidReplyText);
             }
+
             message_lines.push(line);
         }
 
-        self.reply = Some(ErrorReply {
+        self.error_reply = Some(ErrorReply {
             rcode,
             xcode,
             message: message_lines,
@@ -292,57 +309,64 @@ impl ErrorReply {
     // "r x m"
     // "r x m..."
     pub(crate) fn make_error_reply(&self) -> CString {
-        // TODO revise
+        fn fmt_codes(msg: &mut Vec<u8>, rcode: &ReplyCode, xcode: Option<&EnhancedStatusCode>) {
+            write!(msg, "{} ", rcode.as_ref()).unwrap();
 
-        let s = if self.message.len() <= 1 {
-            let mut s = Vec::<u8>::new();
-            write!(s, "{} ", self.rcode.as_ref()).unwrap();
-            if let Some(xcode) = &self.xcode {
-                write!(s, "{}", xcode.as_ref()).unwrap();
+            if let Some(xcode) = xcode {
+                write!(msg, "{}", xcode.as_ref()).unwrap();
             }
-            if let Some(m) = self.message.iter().next() {
+        }
+
+        let msg = match &self.message[..] {
+            [] => {
+                let mut msg = Vec::new();
+                fmt_codes(&mut msg, &self.rcode, self.xcode.as_ref());
+                msg
+            }
+            [text] => {
+                let mut msg = Vec::new();
+                fmt_codes(&mut msg, &self.rcode, self.xcode.as_ref());
                 if self.xcode.is_some() {
-                    write!(s, " ").unwrap();
+                    write!(msg, " ").unwrap();
                 }
-                s.write_all(m.as_bytes()).unwrap();
+                msg.write_all(text.as_bytes()).unwrap();
+                msg
             }
-            s
-        } else {
-            match &self.message[..] {
-                [] | [_] => unreachable!(),
-                [xs @ .., x] => {
-                    let rcode = &self.rcode;
-                    let xcode = self.xcode.as_ref().map_or(match rcode {
+            [lines @ .., last_line] => {
+                let rcode = &self.rcode;
+                let xcode = self.xcode.as_ref().map_or(
+                    match rcode {
                         ReplyCode::Transient(_) => "4.0.0",
                         ReplyCode::Permanent(_) => "5.0.0",
-                    }, |x| x.as_ref());
+                    },
+                    |c| c.as_ref(),
+                );
 
-                    let mut s = Vec::<u8>::new();
+                let mut msg = Vec::<u8>::new();
 
-                    for x in xs {
-                        write!(s, "{}-{} ", rcode.as_ref(), xcode).unwrap();
-                        s.write_all(x.as_bytes()).unwrap();
-                        s.write_all(b"\r\n").unwrap();
-                    }
-
-                    write!(s, "{} {} ", rcode.as_ref(), xcode).unwrap();
-                    s.write_all(x.as_bytes()).unwrap();
-                    s
+                for line in lines {
+                    write!(msg, "{}-{} ", rcode.as_ref(), xcode).unwrap();
+                    msg.write_all(line.as_bytes()).unwrap();
+                    msg.write_all(b"\r\n").unwrap();
                 }
+
+                write!(msg, "{} {} ", rcode.as_ref(), xcode).unwrap();
+                msg.write_all(last_line.as_bytes()).unwrap();
+                msg
             }
         };
 
-        CString::new(s).expect("ill-formed error reply text")
+        CString::new(msg).expect("ill-formed error reply text")
     }
 }
 
-pub struct EomContextActions {
+pub struct EomActions {
     conn: Connection,
     available_actions: Actions,
 }
 
 #[async_trait]
-impl ContextActions for EomContextActions {
+impl ContextActions for EomActions {
     async fn add_header<'cx, 'k, 'v>(
         &'cx self,
         name: impl IntoCString + Send + 'k,
@@ -353,7 +377,7 @@ impl ContextActions for EomContextActions {
         }
 
         let name = name.into_c_string();
-        if name.to_bytes().is_empty() {
+        if name.as_bytes().is_empty() {
             return Err(ContextError::InvalidReply);
         }
 
@@ -373,13 +397,12 @@ impl ContextActions for EomContextActions {
         if !self.available_actions.contains(Actions::ADDHDRS) {
             return Err(ContextError::ActionUnavailable);
         }
-
         if index < 0 {
             return Err(ContextError::InvalidReply);
         }
 
         let name = name.into_c_string();
-        if name.to_bytes().is_empty() {
+        if name.as_bytes().is_empty() {
             return Err(ContextError::InvalidReply);
         }
 
@@ -399,13 +422,12 @@ impl ContextActions for EomContextActions {
         if !self.available_actions.contains(Actions::CHGHDRS) {
             return Err(ContextError::ActionUnavailable);
         }
-
-        let name = name.into_c_string();
-        if name.to_bytes().is_empty() {
+        if index < 0 {
             return Err(ContextError::InvalidReply);
         }
 
-        if index < 0 {
+        let name = name.into_c_string();
+        if name.as_bytes().is_empty() {
             return Err(ContextError::InvalidReply);
         }
 
@@ -426,11 +448,11 @@ impl ContextActions for EomContextActions {
         }
 
         let mail = mail.into_c_string();
-        if mail.to_bytes().is_empty() {
+        if mail.as_bytes().is_empty() {
             return Err(ContextError::InvalidReply);
         }
 
-        let args = args.map(|v| v.into_c_string());
+        let args = args.map(|a| a.into_c_string());
 
         self.conn.write_reply(Reply::ChangeSender { mail, args }).await?;
 
@@ -446,7 +468,7 @@ impl ContextActions for EomContextActions {
         }
 
         let rcpt = rcpt.into_c_string();
-        if rcpt.to_bytes().is_empty() {
+        if rcpt.as_bytes().is_empty() {
             return Err(ContextError::InvalidReply);
         }
 
@@ -465,11 +487,11 @@ impl ContextActions for EomContextActions {
         }
 
         let rcpt = rcpt.into_c_string();
-        if rcpt.to_bytes().is_empty() {
+        if rcpt.as_bytes().is_empty() {
             return Err(ContextError::InvalidReply);
         }
 
-        let args = args.map(|v| v.into_c_string());
+        let args = args.map(|a| a.into_c_string());
 
         self.conn.write_reply(Reply::AddRcptExt { rcpt, args }).await?;
 
@@ -485,7 +507,7 @@ impl ContextActions for EomContextActions {
         }
 
         let rcpt = rcpt.into_c_string();
-        if rcpt.to_bytes().is_empty() {
+        if rcpt.as_bytes().is_empty() {
             return Err(ContextError::InvalidReply);
         }
 
@@ -499,24 +521,18 @@ impl ContextActions for EomContextActions {
             return Err(ContextError::ActionUnavailable);
         }
 
-        // TODO revisit
-
         if chunk.is_empty() {
-            let reply = Reply::ReplaceBody { chunk: vec![] };
-
-            self.conn.write_reply(reply).await?;
+            self.conn.write_reply(Reply::ReplaceBody { chunk: vec![] }).await?;
 
             return Ok(());
         }
 
-        let milter_chunk_size = 65535;
+        const CHUNK_SIZE: usize = 65535;
 
-        let chunks = chunk.chunks(milter_chunk_size);
+        for c in chunk.chunks(CHUNK_SIZE) {
+            let chunk = c.to_vec();
 
-        for c in chunks {
-            let reply = Reply::ReplaceBody { chunk: c.to_vec() };
-
-            self.conn.write_reply(reply).await?;
+            self.conn.write_reply(Reply::ReplaceBody { chunk }).await?;
         }
 
         Ok(())
@@ -537,7 +553,7 @@ impl ContextActions for EomContextActions {
         }
 
         let reason = reason.into_c_string();
-        if reason.to_bytes().is_empty() {
+        if reason.as_bytes().is_empty() {
             return Err(ContextError::InvalidReply);
         }
 
@@ -572,14 +588,15 @@ impl AsRef<str> for ReplyCode {
     }
 }
 
+// Parse logic follows stricter RFC 5321, rather than libmilter impl.
 impl FromStr for ReplyCode {
     type Err = ParseStatusCodeError;
 
     fn from_str(s: &str) -> Result<Self, Self::Err> {
         match s.as_bytes() {
             [x, y, z]
-                if matches!(x, b'4' | b'5')
-                    && matches!(y, b'0'..=b'9')
+                if matches!(x, b'4'..=b'5')
+                    && matches!(y, b'0'..=b'5')
                     && matches!(z, b'0'..=b'9') =>
             {
                 Ok(match x {
@@ -607,6 +624,7 @@ impl AsRef<str> for EnhancedStatusCode {
     }
 }
 
+// Parse logic follows stricter RFC 3463, rather than libmilter impl.
 impl FromStr for EnhancedStatusCode {
     type Err = ParseStatusCodeError;
 
