@@ -1,24 +1,27 @@
 use crate::{
     callbacks::{Callbacks, Status},
-    command::{
-        Actions, BodyPayload, CommandError, CommandFrame, CommandKind, ConnInfoPayload,
-        EnvAddrPayload, HeaderPayload, HeloPayload, MacroPayload, OptNegPayload, ProtoOpts,
-        UnknownPayload,
-    },
     config::Config,
     connection::Connection,
-    context::{Context, EomContext, NegotiateContext},
+    context::{Context, EomContext, NegotiateContext, ReplyCode},
     macros::Stage,
-    reply::Reply,
+    message::{
+        commands::{
+            BodyPayload, CommandError, CommandKind, CommandMessage, ConnInfoPayload,
+            EnvAddrPayload, HeaderPayload, HeloPayload, MacroPayload, OptNegPayload,
+            UnknownPayload,
+        },
+        replies::Reply,
+        Version, MILTER_VERSION,
+    },
+    proto_util::{Actions, ProtoOpts},
 };
 use bytes::Bytes;
 use std::{
     collections::HashMap,
-    convert::TryFrom,
     error::Error,
     ffi::CString,
     fmt::{self, Display, Formatter},
-    io, mem,
+    io,
     sync::Arc,
 };
 use tokio::{
@@ -26,7 +29,7 @@ use tokio::{
     select,
     sync::watch,
 };
-use tracing::{debug, trace};
+use tracing::{debug, trace, warn};
 
 #[derive(Clone, Copy, Debug, Eq, Hash, PartialEq)]
 pub enum State {
@@ -54,8 +57,7 @@ impl State {
             Init, Opts, Conn, Helo, Mail, Rcpt, Data, Header, Eoh, Body, Eom, Quit, Abort, Unknown,
             QuitNc,
         ]
-        .iter()
-        .copied()
+        .into_iter()
     }
 
     fn is_mail_transaction(&self) -> bool {
@@ -63,17 +65,17 @@ impl State {
         matches!(self, Mail | Rcpt | Data | Header | Eoh | Body)
     }
 
-    fn can_reach(&self, target: Self, popts: &ProtoOpts) -> bool {
+    fn can_reach(&self, target: Self, opts: ProtoOpts) -> bool {
         let can_be_skipped = |s: &Self| match s {
-            Self::Conn => popts.contains(ProtoOpts::NOCONNECT),
-            Self::Helo => popts.contains(ProtoOpts::NOHELO),
-            Self::Mail => popts.contains(ProtoOpts::NOMAIL),
-            Self::Rcpt => popts.contains(ProtoOpts::NORCPT),
-            Self::Header => popts.contains(ProtoOpts::NOHDRS),
-            Self::Eoh => popts.contains(ProtoOpts::NOEOH),
-            Self::Body => popts.contains(ProtoOpts::NOBODY),
-            Self::Data => popts.contains(ProtoOpts::NODATA),
-            Self::Unknown => popts.contains(ProtoOpts::NOUNKNOWN),
+            Self::Conn => opts.contains(ProtoOpts::NOCONNECT),
+            Self::Helo => opts.contains(ProtoOpts::NOHELO),
+            Self::Mail => opts.contains(ProtoOpts::NOMAIL),
+            Self::Rcpt => opts.contains(ProtoOpts::NORCPT),
+            Self::Header => opts.contains(ProtoOpts::NOHDRS),
+            Self::Eoh => opts.contains(ProtoOpts::NOEOH),
+            Self::Body => opts.contains(ProtoOpts::NOBODY),
+            Self::Data => opts.contains(ProtoOpts::NODATA),
+            Self::Unknown => opts.contains(ProtoOpts::NOUNKNOWN),
             _ => false,
         };
 
@@ -118,18 +120,18 @@ impl State {
     }
 }
 
-fn callbacks_to_popts<T: Send>(callbacks: &Callbacks<T>) -> ProtoOpts {
-    let mut popts = ProtoOpts::empty();
-    popts.set(ProtoOpts::NOCONNECT, callbacks.connect.is_none());
-    popts.set(ProtoOpts::NOHELO, callbacks.helo.is_none());
-    popts.set(ProtoOpts::NOMAIL, callbacks.mail.is_none());
-    popts.set(ProtoOpts::NORCPT, callbacks.rcpt.is_none());
-    popts.set(ProtoOpts::NOHDRS, callbacks.header.is_none());
-    popts.set(ProtoOpts::NOEOH, callbacks.eoh.is_none());
-    popts.set(ProtoOpts::NOBODY, callbacks.body.is_none());
-    popts.set(ProtoOpts::NODATA, callbacks.data.is_none());
-    popts.set(ProtoOpts::NOUNKNOWN, callbacks.unknown.is_none());
-    popts
+fn opts_from_callbacks<T: Send>(callbacks: &Callbacks<T>) -> ProtoOpts {
+    let mut opts = ProtoOpts::empty();
+    opts.set(ProtoOpts::NOCONNECT, callbacks.connect.is_none());
+    opts.set(ProtoOpts::NOHELO, callbacks.helo.is_none());
+    opts.set(ProtoOpts::NOMAIL, callbacks.mail.is_none());
+    opts.set(ProtoOpts::NORCPT, callbacks.rcpt.is_none());
+    opts.set(ProtoOpts::NOHDRS, callbacks.header.is_none());
+    opts.set(ProtoOpts::NOEOH, callbacks.eoh.is_none());
+    opts.set(ProtoOpts::NOBODY, callbacks.body.is_none());
+    opts.set(ProtoOpts::NODATA, callbacks.data.is_none());
+    opts.set(ProtoOpts::NOUNKNOWN, callbacks.unknown.is_none());
+    opts
 }
 
 pub type SessionResult = Result<(), SessionError>;
@@ -139,12 +141,13 @@ pub type SessionResult = Result<(), SessionError>;
 pub enum SessionError {
     Io(io::Error),
     ParseCommand(CommandError),
-    ConnectionClosed,    // unexpected closing of connection mid-conversation
+    ProtocolVersionNotSupported,
     RequestedActionsNotSupported, // requested actions cannot be fulfilled
     RequestedProtoOptsNotSupported,
-    InvalidStatusInNegotiate(Status),
+    InvalidNegotiateStatus(Status),
     BufferEmpty,  // empty buffer not allowed
     UnknownCommand,
+    Shutdown,
     // TODO...
     // TODO timeout?
 }
@@ -172,56 +175,47 @@ impl From<CommandError> for SessionError {
 pub struct Session<T: Send> {
     conn: Connection,
 
-    context: Context<T>,
-
-    callbacks: Arc<Callbacks<T>>,
+    state: State,
 
     shutdown: watch::Receiver<bool>,
 
-    state: State,
-
-    // TODO negotiation
-
-    initial_actions: Actions,
-    initial_popts: ProtoOpts,
+    callbacks: Arc<Callbacks<T>>,
 
+    context: Context<T>,
     actions: Actions,
-    popts: ProtoOpts,
+    opts: ProtoOpts,
 }
 
 // TODO initialization: note that with QuitNc command, this session can again
 // begin at the negotiate stage: ie need to make sure all fields are reset to
-// initial values
+// initial values?
 impl<T: Send> Session<T> {
     pub fn new<S>(
         stream: S,
         shutdown: watch::Receiver<bool>,
         callbacks: Arc<Callbacks<T>>,
-        config: Arc<Config>,
+        config: &Config,
     ) -> Self
     where
         S: AsyncRead + AsyncWrite + Unpin + Send + 'static,
     {
-        let initial_actions = config.actions;
-        let conn_timeout = config.connection_timeout;
+        let conn = Connection::new(stream, config.connection_timeout);
 
-        let initial_popts = callbacks_to_popts(&callbacks);
-        let conn = Connection::new(stream, conn_timeout);
+        let actions = config.actions;
+        let opts = opts_from_callbacks(&callbacks);
 
         Self {
             conn,
 
-            context: Context::new(),
-
-            callbacks,
-            shutdown,
-
             state: State::Init,
 
-            initial_actions,
-            initial_popts,
-            actions: initial_actions,
-            popts: initial_popts,
+            shutdown,
+
+            callbacks,
+
+            context: Context::new(),
+            actions,
+            opts,
         }
     }
 
@@ -231,12 +225,17 @@ impl<T: Send> Session<T> {
     pub async fn process_commands(mut self) -> SessionResult {
         let result = self.process().await;
 
-        if result.is_err() {
-            if self.state.is_mail_transaction() {
-                if let Some(abort) = &self.callbacks.abort {
-                    let _ = abort(&mut self.context).await;
-                }
+        if result.is_err() && self.state.is_mail_transaction() {
+            if let Some(abort) = &self.callbacks.abort {
+                let _ = abort(&mut self.context).await;
             }
+        }
+
+        // We exited the loop in a different state than Quit: for example if the
+        // quit command was received in the middle of a conversation, or on
+        // error, or when the session was shut down. Call the close callback.
+
+        if self.state != State::Quit {
             if let Some(close) = &self.callbacks.close {
                 let _ = close(&mut self.context).await;
             }
@@ -249,39 +248,34 @@ impl<T: Send> Session<T> {
         // For ever process this session’s connection by handling commands until
         // the quit command is received, or the shutdown signal is received.
 
-        while self.state != State::Quit && !*self.shutdown.borrow() {
-            // Read the next frame from the connection. Always also check if the
-            // shutdown flag was touched, and reread its value, else proceed
-            // with handling the command.
+        while self.state != State::Quit {
+            if *self.shutdown.borrow() {
+                return Err(SessionError::Shutdown);
+            }
 
-            let frame = select! {
-                frame = self.conn.read_frame() => frame?,
+            // Read the next message from the connection. Always also check if
+            // the shutdown flag was touched, and restart loop to reread its
+            // value, else proceed with handling the command.
+
+            let msg = select! {
+                msg = self.conn.read_message() => msg?,
                 _ = self.shutdown.changed() => continue,
             };
 
-            // TODO Note that if read_frame exits regularly (peer closes),
-            // then still need to call abort and close callbacks ...
-            // This is actually an error condition: while expecting a
-            // new command, the client closed the connection -- w/o
-            // issuing the quit/close command.
-            let frame = frame.ok_or(SessionError::ConnectionClosed)?;
+            // At this point we need a `CommandMessage`: a valid command, but
+            // not yet fully parsed: the parsing should only happen once it is
+            // certain that a callback will be called.
 
-            trace!(?frame, "got next frame");
-
-            // At this point we need a 'CommandFrame': a valid command frame,
-            // but not yet fully parsed: the parsing should only happen once it
-            // is determined that a callback will be called.
-
-            let frame = CommandFrame::try_from(frame)
+            let msg = CommandMessage::try_from(msg)
                 .map_err(|_| SessionError::UnknownCommand)?;
 
-            let cmd = frame.kind;
+            let cmd = msg.kind;
 
-            // Check if desired state change from the CommandFrame is feasible,
+            // Check if desired state change from the CommandMessage is feasible
             // taking into account the current set of protocol options.
 
             if let Some(next_state) = cmd.as_state() {
-                if !self.state.can_reach(next_state, &self.popts) {
+                if !self.state.can_reach(next_state, self.opts) {
                     // When the desired state cannot be reached, abort the
                     // current mail transaction, and retry starting from the
                     // HELO state.
@@ -292,12 +286,13 @@ impl<T: Send> Session<T> {
                         }
                     }
 
-                    self.set_state(State::Helo);
+                    self.state = State::Helo;
 
-                    if !self.state.can_reach(next_state, &self.popts) {
+                    if !self.state.can_reach(next_state, self.opts) {
                         // The desired state cannot be reached; but if it's a
                         // quit anyway then exit regularly. Else, *ignore* the
                         // invalid command.
+
                         if next_state == State::Quit {
                             break;
                         } else {
@@ -312,51 +307,33 @@ impl<T: Send> Session<T> {
             // buffer is, too, then transition to the new state, and handle the
             // command.
 
-            ensure_buffer_present(cmd, &frame.buffer)?;
+            ensure_buffer_present(cmd, &msg.buffer)?;
 
             if let Some(next_state) = cmd.as_state() {
-                self.set_state(next_state);
+                trace!(state = ?next_state, "transitioned to next state");
+                self.state = next_state;
             }
 
-            self.handle_command(frame).await?;
-        }
-
-        // TODO still call abort? may either be triggered by `?` and returning
-        // early, or by receiving the shutdown signal and landing here!
-
-        // We exited the loop in a different state than quit: for example if the
-        // quit command was received in the middle of a conversation. Call the
-        // close callback.
-
-        if self.state != State::Quit {
-            if let Some(close) = &self.callbacks.close {
-                let _ = close(&mut self.context).await;
-            }
+            self.handle_command(msg).await?;
         }
 
         Ok(())
     }
 
-    fn set_state(&mut self, state: State) {
-        // TODO however, note distinction curstate vs ctx_state?
-        trace!(?state, "transitioned to new state");
-        self.state = state;
-    }
-
-    async fn handle_command(&mut self, frame: CommandFrame) -> SessionResult {
+    async fn handle_command(&mut self, msg: CommandMessage) -> SessionResult {
         // Handling a command depends on the command kind. The command buffer is
         // only parsed (and parsing can only fail) if a callback is available.
-        // The exception are `OptNeg` and `DefMacros` where the buffer is parsed
-        // unconditionally and special handling applies.
+        // The exceptions are `OptNeg` and `DefMacros` where the buffer is
+        // parsed unconditionally and special handling applies.
 
-        let status = match frame.kind {
+        let status = match msg.kind {
             CommandKind::OptNeg => {
-                self.handle_opt_neg_command(frame.buffer).await?;
+                self.handle_opt_neg_command(msg.buffer).await?;
 
                 return Ok(());
             }
-            CommandKind::DefMacro => {
-                self.handle_def_macro_command(frame.buffer);
+            CommandKind::DefMacros => {
+                self.handle_def_macros_command(msg.buffer);
 
                 return Ok(());
             }
@@ -367,7 +344,7 @@ impl<T: Send> Session<T> {
                     let ConnInfoPayload {
                         hostname,
                         socket_info,
-                    } = ConnInfoPayload::parse_buffer(frame.buffer)?;
+                    } = ConnInfoPayload::parse_buffer(msg.buffer)?;
                     connect(&mut self.context, hostname, socket_info).await
                 } else {
                     Status::Continue
@@ -377,7 +354,7 @@ impl<T: Send> Session<T> {
                 self.context.clear_macros_after(Stage::Helo);
 
                 if let Some(helo) = &self.callbacks.helo {
-                    let HeloPayload { hostname } = HeloPayload::parse_buffer(frame.buffer)?;
+                    let HeloPayload { hostname } = HeloPayload::parse_buffer(msg.buffer)?;
                     helo(&mut self.context, hostname).await
                 } else {
                     Status::Continue
@@ -387,7 +364,7 @@ impl<T: Send> Session<T> {
                 self.context.clear_macros_after(Stage::Mail);
 
                 if let Some(mail) = &self.callbacks.mail {
-                    let EnvAddrPayload { args } = EnvAddrPayload::parse_buffer(frame.buffer)?;
+                    let EnvAddrPayload { args } = EnvAddrPayload::parse_buffer(msg.buffer)?;
                     mail(&mut self.context, args).await
                 } else {
                     Status::Continue
@@ -397,7 +374,7 @@ impl<T: Send> Session<T> {
                 self.context.clear_macros_after(Stage::Rcpt);
 
                 if let Some(rcpt) = &self.callbacks.rcpt {
-                    let EnvAddrPayload { args } = EnvAddrPayload::parse_buffer(frame.buffer)?;
+                    let EnvAddrPayload { args } = EnvAddrPayload::parse_buffer(msg.buffer)?;
                     rcpt(&mut self.context, args).await
                 } else {
                     Status::Continue
@@ -412,7 +389,7 @@ impl<T: Send> Session<T> {
             }
             CommandKind::Header => {
                 if let Some(header) = &self.callbacks.header {
-                    let HeaderPayload { name, value } = HeaderPayload::parse_buffer(frame.buffer)?;
+                    let HeaderPayload { name, value } = HeaderPayload::parse_buffer(msg.buffer)?;
                     header(&mut self.context, name, value).await
                 } else {
                     Status::Continue
@@ -427,7 +404,7 @@ impl<T: Send> Session<T> {
             }
             CommandKind::BodyChunk => {
                 if let Some(body) = &self.callbacks.body {
-                    let BodyPayload { chunk } = BodyPayload::parse_buffer(frame.buffer)?;
+                    let BodyPayload { chunk } = BodyPayload::parse_buffer(msg.buffer)?;
                     body(&mut self.context, chunk).await
                 } else {
                     Status::Continue
@@ -437,8 +414,8 @@ impl<T: Send> Session<T> {
                 let mut status = Status::Continue;
 
                 if let Some(body) = &self.callbacks.body {
-                    if !frame.buffer.is_empty() {
-                        let BodyPayload { chunk } = BodyPayload::parse_buffer(frame.buffer)?;
+                    if !msg.buffer.is_empty() {
+                        let BodyPayload { chunk } = BodyPayload::parse_buffer(msg.buffer)?;
                         status = body(&mut self.context, chunk).await;
 
                         if status != Status::Continue {
@@ -449,7 +426,7 @@ impl<T: Send> Session<T> {
 
                 if status == Status::Continue {
                     if let Some(eom) = &self.callbacks.eom {
-                        let mut ctx = EomContext::new(
+                        let mut cx = EomContext::new(
                             self.conn.clone(),
                             self.context.data.take(),
                             self.context.macros.duplicate(),
@@ -457,10 +434,9 @@ impl<T: Send> Session<T> {
                             self.actions,
                         );
 
-                        status = eom(&mut ctx).await;
+                        status = eom(&mut cx).await;
 
-                        // TODO restore context:
-                        self.context.restore_from_eom(ctx);
+                        self.context.restore(cx.data, cx.reply);
                     }
                 }
 
@@ -468,25 +444,23 @@ impl<T: Send> Session<T> {
             }
             CommandKind::Abort => {
                 if let Some(abort) = &self.callbacks.abort {
-                    abort(&mut self.context).await
-                } else {
-                    Status::Continue
+                    let _ = abort(&mut self.context).await;
                 }
+
+                return Ok(());
             }
             CommandKind::Quit | CommandKind::QuitNc => {
-                let status = if let Some(close) = &self.callbacks.close {
-                    close(&mut self.context).await
-                } else {
-                    Status::Continue
-                };
+                if let Some(close) = &self.callbacks.close {
+                    let _ = close(&mut self.context).await;
+                }
 
                 self.context.clear_macros();
 
-                status
+                return Ok(());
             }
             CommandKind::Unknown => {
                 if let Some(unknown) = &self.callbacks.unknown {
-                    let UnknownPayload { arg } = UnknownPayload::parse_buffer(frame.buffer)?;
+                    let UnknownPayload { arg } = UnknownPayload::parse_buffer(msg.buffer)?;
                     unknown(&mut self.context, arg).await
                 } else {
                     Status::Continue
@@ -503,192 +477,190 @@ impl<T: Send> Session<T> {
             || matches!(status, Status::Reject | Status::Discard | Status::Tempfail)
                 && !matches!(self.state, State::Rcpt | State::Unknown)
         {
-            self.set_state(State::Helo);
+            self.state = State::Helo;
         }
-        // TODO abort? not possible from callback I think -- check
-        // can *only* happen in bodyend when sending reply after body callback (?)
-        // and also when negotiate aborts, eg cannot satisfy milter reqs
 
         Ok(())
     }
 
     async fn handle_opt_neg_command(&mut self, buffer: Bytes) -> SessionResult {
+        // At this stage there are usually no macros to be cleared, however,
+        // note that macros can be sent at any time, even before negotiate.
         self.context.clear_macros();
 
         let OptNegPayload {
-            version,
-            actions: action_flags,
-            popts: protocol_flags,
+            version: mta_version,
+            actions: mut mta_actions,
+            opts: mut mta_opts,
         } = OptNegPayload::parse_buffer(buffer)?;
 
-        // TODO update session object with info
-
-        let mta_version = version;
-
-        // `mta_actions` are received in negotiate, and checked for
-        // compliance, but not *used* anywhere later. However they are
-        // used for `actions`.
-        let mut mta_actions = action_flags;
-        let mut mta_protocol_opts = protocol_flags;
-
-        // clear macros > neg
-
-        assert!(mta_version >= 6);
+        // Note: different from libmilter. We only support protocol versions 6
+        // (and up). Older versions are not supported.
+        if mta_version < MILTER_VERSION {
+            return Err(SessionError::ProtocolVersionNotSupported);
+        }
 
         if mta_actions.is_empty() {
-            mta_actions = Actions::from_bits_truncate(0xf);
+            mta_actions = Actions::MIN_ACTIONS;
+        }
+        if mta_opts.is_empty() {
+            mta_opts = ProtoOpts::MIN_PROTO_OPTS;
         }
 
-        if mta_protocol_opts.is_empty() {
-            mta_protocol_opts = ProtoOpts::from_bits_truncate(0x3f);
-        }
+        // TODO FFR max data size?
 
-        self.actions = self.initial_actions;
+        let target_opts;
+        let requested_macros;
 
-        let popts2mta;
-        let mut requested_macros = Default::default();
-
-        // If a callback is available, ...
         if let Some(negotiate) = &self.callbacks.negotiate {
-            // TODO more stuff ...
+            let default_actions = mta_actions;
+            let default_opts = self.opts | (mta_opts & ProtoOpts::SKIP);
 
-            let mut ctx = NegotiateContext::new(
-                // self.context.macros.duplicate(),
+            let mut cx = NegotiateContext::new(
+                self.context.data.take(),
                 self.context.reply.duplicate(),
-                mta_actions,
-                self.initial_popts,
+                default_actions,
+                default_opts,
             );
 
-            let status = negotiate(&mut ctx, mta_actions, mta_protocol_opts).await;
+            let status = negotiate(&mut cx, mta_actions, mta_opts).await;
 
-            let requested_actions = ctx.requested_actions;
-            let requested_popts = ctx.requested_proto_opts;
-            requested_macros = mem::take(&mut ctx.requested_macros);
+            self.context.restore(cx.data, cx.reply);
 
-            // TODO restore context
-            self.context.data = ctx.data;
-            self.context.reply = ctx.reply;
+            requested_macros = cx.requested_macros;
 
             match status {
                 Status::AllOpts => {
                     self.actions = mta_actions;
-                    popts2mta = self.initial_popts;
-                    // ...
+                    target_opts = default_opts;
                 }
                 Status::Continue => {
-                    self.actions = requested_actions;
-                    popts2mta = requested_popts;
-                    // ...
+                    self.actions = cx.requested_actions;
+                    self.opts = cx.requested_opts;
+                    target_opts = cx.requested_opts;
                 }
                 status => {
-                    return Err(SessionError::InvalidStatusInNegotiate(status));
+                    return Err(SessionError::InvalidNegotiateStatus(status));
                 }
             }
 
-            // TODO ...
+            // Noreply options requested but not supported by the MTA are not
+            // ‘faked’ (unlike in libmilter).
         } else {
-            popts2mta = self.initial_popts;
-
-            // TODO ...
+            target_opts = self.opts;
+            requested_macros = Default::default();
         }
 
-        // TODO else, update session a bit more, check requirements etc. (may error)
-        // especially update stages (some may be disabled even though available)!
-
         if !mta_actions.contains(self.actions) {
             return Err(SessionError::RequestedActionsNotSupported);
         }
-
-        if !mta_protocol_opts.contains(popts2mta) {
+        if !mta_opts.contains(target_opts) {
             return Err(SessionError::RequestedProtoOptsNotSupported);
         }
 
-        // TODO fix_stm
-
-        self.write_optneg_reply(self.actions, popts2mta, requested_macros)
+        self.write_opt_neg_reply(MILTER_VERSION, self.actions, target_opts, requested_macros)
             .await?;
 
         Ok(())
     }
 
-    fn handle_def_macro_command(&mut self, buffer: Bytes) {
-        // Macro definitions are handled leniently. Failures are silent, eg when
-        // the buffer contains unusable data, there are empty keys, or if there
-        // is not an even number of macro keys/values.
-        // TODO though note that some of this isn't really "failure": eg Postfix
-        // will send empty macros ("") which cannot be parsed ...
+    fn handle_def_macros_command(&mut self, buffer: Bytes) {
+        // Note that parsing, and handling of macro definitions from the MTA is
+        // best-effort. The current state does not matter. Failures are silent,
+        // eg when the buffer contains unusable data, or if there is not an even
+        // number of macro keys/values.
+        //
+        // Such ‘failures’ do occur in practice. For example, Postfix will send
+        // an empty macro payload, which cannot be parsed.
 
         let MacroPayload { stage, macros } = match MacroPayload::parse_buffer(buffer) {
             Ok(payload) => payload,
             Err(e) => {
-                // Errors in a macro command are ignored. Resume loop.
-                debug!("invalid macro command received: {}", e);
+                trace!("skipping unrecognized macro command: {}", e);
                 return;
             }
         };
 
-        // TODO Note: convert to str, drop isolated excess value, remove empty keys
-        let pairs = macros
-            .chunks(2)
-            .map(|chunk| match chunk {
-                [k, v] => (k.to_owned(), v.to_owned()),
-                _ => unreachable!(),
-            })
-            .filter(|(k, _)| !k.as_bytes().is_empty())
-            .collect();
+        // Note that the transformation below only retains the first occurrence
+        // of a key, and drops an isolated excess value. These modifications are
+        // silent.
 
-        debug!(?stage, ?pairs, "inserted new macros");
+        let mut entries = HashMap::new();
 
-        self.context.insert_macros(stage, pairs);
+        let mut macros = macros.into_iter();
+        while let (Some(k), Some(v)) = (macros.next(), macros.next()) {
+            entries.entry(k).or_insert(v);
+        }
+
+        trace!(?stage, ?entries, "registered new macro definitions");
+
+        self.context.insert_macros(stage, entries);
     }
 
-    async fn write_optneg_reply(
+    async fn write_opt_neg_reply(
         &mut self,
+        version: Version,
         requested_actions: Actions,
-        requested_popts: ProtoOpts,
+        requested_opts: ProtoOpts,
         requested_macros: HashMap<Stage, CString>,
     ) -> io::Result<()> {
-        let c = Reply::OptNeg {
-            version: 0x6,
-            actions: requested_actions, //Actions::from_bits_truncate(0x1ff),
-            popts: requested_popts,
-            requested_macros,
+        let reply = Reply::OptNeg {
+            version,
+            actions: requested_actions,
+            opts: requested_opts,
+            macros: requested_macros,
         };
 
-        self.conn.write_reply(c).await
+        self.conn.write_reply(reply).await
     }
 
-    async fn write_reply(&mut self, status: Status) -> io::Result<()> {
-        if matches!(self.state, State::Quit | State::QuitNc | State::Abort) {
-            // no reply if regularly received quit or quit_nc command
-            return Ok(());
+    async fn write_reply(&mut self, mut status: Status) -> io::Result<()> {
+        fn needs_noreply(opts: ProtoOpts, state: State) -> bool {
+            match state {
+                State::Conn => opts.contains(ProtoOpts::NR_CONN),
+                State::Helo => opts.contains(ProtoOpts::NR_HELO),
+                State::Mail => opts.contains(ProtoOpts::NR_MAIL),
+                State::Rcpt => opts.contains(ProtoOpts::NR_RCPT),
+                State::Data => opts.contains(ProtoOpts::NR_DATA),
+                State::Unknown => opts.contains(ProtoOpts::NR_UNKN),
+                State::Header => opts.contains(ProtoOpts::NR_HDR),
+                State::Eoh => opts.contains(ProtoOpts::NR_EOH),
+                State::Body => opts.contains(ProtoOpts::NR_BODY),
+                _ => false,
+            }
         }
 
-        fn reject_or_tempfail<T: Send>(session: &mut Session<T>, initial: u8, r: Reply) -> Reply {
-            session
-                .context
-                .reply
-                .take_reply_if_init_eq(initial)
-                .map_or(r, |reply| Reply::ReplyCode { reply })
+        // Warn on milter user error: user should fix milter implementation.
+        // Still, MTA communication is salvaged by fixing status.
+        if needs_noreply(self.opts, self.state) && status != Status::Noreply {
+            warn!("status response Noreply requested but not used");
+            status = Status::Noreply;
         }
 
-        let c = match status {
+        let reply = match status {
             Status::Accept => Reply::Accept,
             Status::Continue => Reply::Continue,
-            Status::Reject => reject_or_tempfail(self, b'5', Reply::Reject),
-            Status::Tempfail => reject_or_tempfail(self, b'4', Reply::Tempfail),
+            Status::Reject => {
+                self.context
+                    .take_reply_if(|r| matches!(r, ReplyCode::Permanent(_)))
+                    .unwrap_or(Reply::Reject)
+            }
+            Status::Tempfail => {
+                self.context
+                    .take_reply_if(|r| matches!(r, ReplyCode::Transient(_)))
+                    .unwrap_or(Reply::Tempfail)
+            }
             Status::Discard => Reply::Discard,
             Status::Skip => Reply::Skip,
             Status::Noreply => {
-                // TODO write continue if not supported
-                return Ok(());
-            }
-            Status::AllOpts => {
+                // ‘Faked’ Noreply (= Continue) not supported: Noreply was
+                // negotiated successfully with MTA.
                 return Ok(());
             }
+            _ => return Ok(()),
         };
 
-        self.conn.write_reply(c).await
+        self.conn.write_reply(reply).await
     }
 }
 
@@ -696,7 +668,7 @@ fn ensure_buffer_present(cmd: CommandKind, buf: &[u8]) -> SessionResult {
     use CommandKind::*;
     if matches!(
         cmd,
-        DefMacro | BodyChunk | ConnInfo | Helo | Header | Mail | OptNeg | Rcpt | Unknown
+        DefMacros | BodyChunk | ConnInfo | Helo | Header | Mail | OptNeg | Rcpt | Unknown
     ) && buf.is_empty()
     {
         return Err(SessionError::BufferEmpty);
@@ -713,7 +685,7 @@ mod tests {
     fn can_reach_ok() {
         use State::*;
 
-        let popts = ProtoOpts::NOHELO
+        let opts = ProtoOpts::NOHELO
             | ProtoOpts::NOMAIL
             | ProtoOpts::NOHDRS
             | ProtoOpts::NOEOH
@@ -721,13 +693,13 @@ mod tests {
             | ProtoOpts::NODATA
             | ProtoOpts::NOUNKNOWN;
 
-        assert!(Init.can_reach(Opts, &popts));
-        assert!(Opts.can_reach(Conn, &popts));
-        assert!(!Conn.can_reach(Conn, &popts));
-        assert!(!Opts.can_reach(Helo, &popts));
-        assert!(Conn.can_reach(Helo, &popts));
-        assert!(Conn.can_reach(Mail, &popts));
-        assert!(Conn.can_reach(Rcpt, &popts));
-        assert!(!Conn.can_reach(Data, &popts));
+        assert!(Init.can_reach(Opts, opts));
+        assert!(Opts.can_reach(Conn, opts));
+        assert!(!Conn.can_reach(Conn, opts));
+        assert!(!Opts.can_reach(Helo, opts));
+        assert!(Conn.can_reach(Helo, opts));
+        assert!(Conn.can_reach(Mail, opts));
+        assert!(Conn.can_reach(Rcpt, opts));
+        assert!(!Conn.can_reach(Data, opts));
     }
 }
